library ieee;
    use ieee.std_logic_1164.all;
    use ieee.numeric_std.all;
    
entity hello_world is
    port
    (
        clk     	    : in std_logic;         -- input clock
		button		    : in std_logic;         -- button on board (PIN_144)
        led     	    : out std_logic;        -- LED on board (PIN_3)
        led2    	    : out std_logic;        -- LED on board (PIN_7)
        led3    	    : out std_logic;        -- LED on board (PIN_9)
        
        txd             : out std_logic;        -- UART out pin (PIN_31)
        rxd             : in std_logic          -- UART in pin (PIN_32)
    );
end entity hello_world;

architecture rtl of hello_world is
	constant CLK_FREQ   	: integer := 50000000;
	
    signal clk200           : std_logic;
    
	signal blink_freq 		: integer range 0 to 255 := 51;
	signal cnt_max 			: unsigned(24 downto 0) := 25ux"1FFFFFF";
    
    -- everything that's needed for blinking
    
	signal cnt      		: unsigned(24 downto 0) := (others => '0');
	signal blink    		: std_logic_vector (2 downto 0) := "000";
    signal old_blink        : std_logic_vector (2 downto 0) := "001";
	signal button_prev		: std_logic := '1';
	signal button_debounced	: std_logic := '1';
	signal mult_result		: unsigned(15 downto 0);
	
    -- the UART signals
    
    signal rx_data          : std_logic_vector (7 downto 0);
    signal rx_busy          : std_logic;
    signal tx_data          : std_logic_vector (7 downto 0);
    signal tx_start         : std_logic;
    signal tx_busy          : std_logic;
    
    -- the UART output 
    
    signal str              : string(1 to 5);
    signal str_index        : integer range 1 to 10 := 1;
    
    -- UART input
    signal rcv_char         : character := character'val(0);
    signal prev_rx_busy     : std_logic := '0';
    
    
    function log2(i : natural) return integer is
        variable temp    : integer := i;
        variable ret_val : integer := 0; 
    begin					
        while temp > 1 loop
            ret_val := ret_val + 1;
            temp    := temp / 2;     
        end loop;
  	    return ret_val;
    end function;
    
    function strrev(s : string) return string is
        variable sr : string(s'low to s'high);
    begin
        for i in s'high to s'low loop
            sr(s'high - i) := s(i);
        end loop;
        return sr;
    end function strrev;
    
    function slv_to_str(slv : std_logic_vector; radix : integer) return string is
        variable r : string(1 to 5) := "     ";
        variable v : integer := to_integer(unsigned(slv));
        variable d : integer := 2 ** slv'high - 1;
        constant nums : string := "0123456789ABCDEF";
    begin
        for j in 1 to log2(to_integer(unsigned(slv))) loop
            r(j) := nums(v rem radix);
            v := v / 10;
        end loop;    
        return strrev(r);
    end function slv_to_str;
    
begin
    i_pll : entity work.pll
        port map
        (
            inclk0 => clk,
            c0 => clk200
        );
        
    -- the button debouncer logic
    
	i_debounce : entity work.debounce
		port map
		(
			clk => clk,
			button => button,
			result => button_debounced
		);
	
    -- ALTERA 8 x 8 hardware multiplier
    
	i_multiplier : entity work.multiplier
		port map
		(
			a => to_unsigned(blink_freq, 8),
			b => x"ff",
			clk => clk,
			aclr => '0',
			result => mult_result
		);

    -- simple UARt
    
    i_uart : entity work.uart
        generic map
        (
            clk_freq => 200000000,
            baudrate => 460800
        )
        port map
        (
            rxd => rxd,
            rx_data => rx_data,
            rx_busy => rx_busy,
            txd => txd,
            tx_data => tx_data,
            tx_start => tx_start,
            tx_busy => tx_busy,
            clk => clk200            
        );
    
	p_doit : process(clk)
	begin
		if rising_edge(clk) then
            
            -- do blinker logic
            
			if cnt = cnt_max then
				cnt <= (others => '0');
				blink(0) <= not blink(0);
				blink(1) <= not blink(1);
			else
				if cnt = cnt_max / 2 then
					blink(1) <= not blink(1);
					blink(2) <= not blink(2);
				end if;
				cnt <= cnt + 1;
			end if;
            report(slv_to_str(rx_data, 10)) severity warning;
			
            -- react on button press
            
			if button_debounced = '0' and button_prev /= '0' then
				blink_freq <= blink_freq + 51;

				if blink_freq = 255 then
					blink_freq <= 51;
				end if;
			end if;
			button_prev <= button_debounced;
            
            -- output string to UART
            
            if tx_busy /= '1' and blink /= old_blink then
                tx_data <= std_logic_vector(to_unsigned(character'pos(str(str_index)), 8));
                tx_start <= '1';
                str_index <= str_index + 1;
                if str_index > str'length then
                    str_index <= 1;
                    old_blink <= blink;
                end if;
            else
                tx_start <= '0';
            end if;
           
            
            -- receive from UART and deal with key
            
            if rx_busy /= '1' and prev_rx_busy = '1' then
                rcv_char <= character'val(to_integer(unsigned(rx_data)));
                if rcv_char = '-' then
                    blink_freq <= blink_freq + 51;
                    if blink_freq = 255 then
                        blink_freq <= 51;
                    end if;
                    cnt <= (others => '0');
                elsif rcv_char = '+' then
                    blink_freq <= blink_freq - 51;
                    if blink_freq = 51 then
                        blink_freq <= 255;
                    end if;
                    cnt <= (others => '0');
                end if;
            end if;
            prev_rx_busy <= rx_busy;
		end if;
    end process p_doit;

    cnt_max <= mult_result & "000000000";
    
    -- prepare uart output
    str(1) <= '-' when blink(0) = '1' else 'O';
    str(2) <= '-' when blink(1) = '1' else 'O';
    str(3) <= '-' when blink(2) = '1' else 'O';
    str(4) <= character'val(13);        -- CR
    str(5) <= character'val(10);        -- LF
    
    -- prepare LED output
	led <= blink(0);
	led2 <= blink(1);
	led3 <= blink(2);
end rtl;
